DATASET_MANIFEST_FILE = ".dataset_metadata/.gantry_manifest.jsonl"  # manifest file
DATASET_HEAD_FILE = ".dataset_metadata/HEAD"  # current head commit info
GANTRY_FOLDER = ".dataset_metadata"
BACKUP_SUFFIX = "_backup"
NEW_SUFFIX = "_new"

FILE_NAME = "file_name"
NEW_FILES = "new_files"
MODIFIED_FILES = "modified_files"
DELETED_FILES = "deleted_files"
UNCHANGED_FILES = "unchanged_files"
SHA256 = "sha256"
URL = "url"


METADATA_S3_FILE_VERSION = "metadata_s3_file_version"


FILE_PATH = "file_path"
OBJ_KEY = "obj_key"
VERSION_ID = "version_id"
